from typing import Iterable, List, Optional

from benchling_api_client.v2.beta.api.worklists import (
    create_worklist,
    create_worklist_item,
    delete_worklist,
    get_worklist,
    list_worklists,
    update_worklist,
)
from benchling_api_client.v2.beta.models.list_worklists_sort import ListWorklistsSort
from benchling_api_client.v2.beta.models.worklist import Worklist
from benchling_api_client.v2.beta.models.worklist_create import WorklistCreate
from benchling_api_client.v2.beta.models.worklist_item import WorklistItem
from benchling_api_client.v2.beta.models.worklist_item_create import WorklistItemCreate
from benchling_api_client.v2.beta.models.worklist_type import WorklistType
from benchling_api_client.v2.beta.models.worklist_update import WorklistUpdate
from benchling_api_client.v2.beta.models.worklists_paginated_list import WorklistsPaginatedList
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaWorklistService(BaseService):
    """
    V2-Beta Worklists.

    Worklists are a convenient way to organize items for bulk actions, and are complementary to folders and
    projects.

    See https://benchling.com/api/v2-beta/reference#/Worklists
    """

    @api_method
    def _worklists_page(
        self,
        *,
        sort: Optional[ListWorklistsSort] = None,
        modified_at: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        worklist_type: Optional[WorklistType] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[WorklistsPaginatedList]:
        return list_worklists.sync_detailed(  # type: ignore
            client=self.client,
            sort=none_as_unset(sort),
            modified_at=none_as_unset(modified_at),
            ids=none_as_unset(optional_array_query_param(ids)),
            worklist_type=none_as_unset(worklist_type),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )

    def list(
        self,
        *,
        sort: Optional[ListWorklistsSort] = None,
        modified_at: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        worklist_type: Optional[WorklistType] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[Worklist]:
        """
        List worklists.

        Individual items within a worklist are summarized.

        See https://benchling.com/api/v2-beta/reference#/Worklists/listWorklists
        """

        def api_call(next_token: NextToken) -> Response[WorklistsPaginatedList]:
            return self._worklists_page(
                sort=sort,
                modified_at=modified_at,
                ids=ids,
                worklist_type=worklist_type,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: WorklistsPaginatedList) -> Optional[List[Worklist]]:
            return body.worklists

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, worklist: WorklistCreate) -> Worklist:
        """
        Create a worklist.

        See https://benchling.com/api/v2-beta/reference#/Worklists/createWorklist
        """
        response = create_worklist.sync_detailed(client=self.client, json_body=worklist)
        return model_from_detailed(response)

    @api_method
    def delete(self, worklist_id: str) -> None:
        """
        Permanently deletes a worklist.

        See https://benchling.com/api/v2-beta/reference#/Worklists/deleteWorklist
        """
        response = delete_worklist.sync_detailed(client=self.client, worklist_id=worklist_id)
        raise_for_status(response)

    @api_method
    def get_by_id(self, worklist_id: str) -> Worklist:
        """
        Get a worklist by ID.

        See https://benchling.com/api/v2-beta/reference#/Worklists/getWorklist
        """
        response = get_worklist.sync_detailed(client=self.client, worklist_id=worklist_id)
        return model_from_detailed(response)

    @api_method
    def update(self, worklist_id: str, worklist_update: WorklistUpdate) -> Worklist:
        """
        Update a worklist.

        See https://benchling.com/api/v2-beta/reference#/Worklists/updateWorklist
        """
        response = update_worklist.sync_detailed(
            client=self.client, worklist_id=worklist_id, json_body=worklist_update
        )
        return model_from_detailed(response)

    @api_method
    def append_item(self, worklist_id, worklist_item_create: WorklistItemCreate) -> WorklistItem:
        """
        Append an item to the end of a worklist if the item is not already present in the worklist.

        See https://benchling.com/api/v2-beta/reference#/Worklists/createWorklistItem
        """
        response = create_worklist_item.sync_detailed(
            client=self.client, worklist_id=worklist_id, json_body=worklist_item_create
        )
        return model_from_detailed(response)



from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.apps import (
    archive_benchling_apps,
    create_benchling_app,
    get_benchling_app_by_id,
    list_benchling_apps,
    patch_benchling_app,
    unarchive_benchling_apps,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import (
    BenchlingApp,
    BenchlingAppCreate,
    BenchlingAppsArchivalChange,
    BenchlingAppsArchive,
    BenchlingAppsArchiveReason,
    BenchlingAppsPaginatedList,
    BenchlingAppsUnarchive,
    BenchlingAppUpdate,
    ListBenchlingAppsSort,
)
from benchling_sdk.services.v2.base_service import BaseService


class AppService(BaseService):
    """
    Apps.

    Create and manage Apps on your tenant.

    See https://benchling.com/api/reference#/Apps
    """

    @api_method
    def _apps_page(
        self,
        *,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        sort: Optional[ListBenchlingAppsSort] = None,
        ids: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        namesany_of: Optional[Iterable[str]] = None,
        namesany_ofcase_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        member_of: Optional[Iterable[str]] = None,
        admin_of: Optional[Iterable[str]] = None,
    ) -> Response[BenchlingAppsPaginatedList]:
        response = list_benchling_apps.sync_detailed(
            client=self.client,
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            sort=none_as_unset(sort),
            ids=none_as_unset(optional_array_query_param(ids)),
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            namesany_of=none_as_unset(optional_array_query_param(namesany_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(namesany_ofcase_sensitive)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            member_of=none_as_unset(optional_array_query_param(member_of)),
            admin_of=none_as_unset(optional_array_query_param(admin_of)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_apps(
        self,
        *,
        page_size: Optional[int] = 50,
        sort: Optional[ListBenchlingAppsSort] = None,
        ids: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        namesany_of: Optional[Iterable[str]] = None,
        namesany_ofcase_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[str] = None,
        member_of: Optional[str] = None,
        admin_of: Optional[str] = None,
    ) -> PageIterator[BenchlingApp]:
        """
        List Apps.

        See https://benchling.com/api/reference#/Apps/listBenchlingApps
        """

        def api_call(next_token: NextToken) -> Response[BenchlingAppsPaginatedList]:
            return self._apps_page(
                page_size=page_size,
                sort=sort,
                ids=ids,
                modified_at=modified_at,
                name=name,
                name_includes=name_includes,
                namesany_of=namesany_of,
                namesany_ofcase_sensitive=namesany_ofcase_sensitive,
                creator_ids=creator_ids,
                member_of=member_of,
                admin_of=admin_of,
                next_token=next_token,
            )

        def results_extractor(body: BenchlingAppsPaginatedList) -> Optional[List[BenchlingApp]]:
            return body.apps

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_by_id(self, app_id: str) -> BenchlingApp:
        """
        Get an App by ID.

        See https://benchling.com/api/reference#/Apps/getBenchlingAppByID
        """
        response = get_benchling_app_by_id.sync_detailed(client=self.client, app_id=app_id)
        return model_from_detailed(response)

    @api_method
    def create(self, app: BenchlingAppCreate) -> BenchlingApp:
        """
        Create an App.

        See https://benchling.com/api/reference#/Apps/createBenchlingApp
        """
        response = create_benchling_app.sync_detailed(client=self.client, json_body=app)
        return model_from_detailed(response)

    @api_method
    def update(self, app_id: str, app: BenchlingAppUpdate) -> BenchlingApp:
        """
        Update an App's metadata.

        See https://benchling.com/api/reference#/Apps/patchBenchlingApp
        """
        response = patch_benchling_app.sync_detailed(client=self.client, app_id=app_id, json_body=app)
        return model_from_detailed(response)

    @api_method
    def archive(
        self, app_ids: Iterable[str], reason: BenchlingAppsArchiveReason
    ) -> BenchlingAppsArchivalChange:
        """
        Archive Apps.

        See https://benchling.com/api/reference#/Apps/archiveBenchlingApps
        """
        archive_request = BenchlingAppsArchive(app_ids=list(app_ids), reason=reason)
        response = archive_benchling_apps.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, app_ids: Iterable[str]) -> BenchlingAppsArchivalChange:
        """
        Unarchive Apps.

        See https://benchling.com/api/reference#/Apps/unarchiveBenchlingApps
        """
        unarchive_request = BenchlingAppsUnarchive(app_ids=list(app_ids))
        response = unarchive_benchling_apps.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)



from benchling_api_client.v2.beta.api.apps import get_app_configuration_by_app_id
from benchling_api_client.v2.beta.models.benchling_app_configuration import BenchlingAppConfiguration

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaAppService(BaseService):
    """
    V2-Beta Apps.

    Create and manage Apps on your tenant.

    https://benchling.com/api/v2-beta/reference?stability=not-available#/Apps
    """

    @api_method
    def get_configuration_by_app_id(self, app_id: str) -> BenchlingAppConfiguration:
        """
        Get an app's configuration by app id.

        See https://benchling.com/api/v2-beta/reference?stability=la#/Apps/getAppConfigurationByAppID
        """
        response = get_app_configuration_by_app_id.sync_detailed(client=self.client, app_id=app_id)
        return model_from_detailed(response)



import base64

import attr
from benchling_api_client.v2.benchling_client import AuthorizationMethod


@attr.s(auto_attribs=True)
class ApiKeyAuth(AuthorizationMethod):
    """
    API Key Authorization.

    Use in combination with the Benchling() client constructor to be authorized with HTTP Basic Authorization

    :param api_key: The Benchling-provided API key to use.
    """

    api_key: str

    def get_authorization_header(self, *args, **kwargs) -> str:
        """Get content for a HTTP Authorization header."""
        token_encoded = base64.b64encode(f"{self.api_key}:".encode())
        return f"Basic {token_encoded.decode()}"



from benchling_api_client.v2.stable.client import Client

from benchling_sdk.helpers.client_helpers import v2_beta_client
from benchling_sdk.helpers.retry_helpers import RetryStrategy
from benchling_sdk.services.v2.base_service import BaseService
from benchling_sdk.services.v2.beta.v2_beta_aa_sequence_service import V2BetaAaSequenceService
from benchling_sdk.services.v2.beta.v2_beta_app_service import V2BetaAppService
from benchling_sdk.services.v2.beta.v2_beta_custom_entity_service import V2BetaCustomEntityService
from benchling_sdk.services.v2.beta.v2_beta_dna_oligo_service import V2BetaDnaOligoService
from benchling_sdk.services.v2.beta.v2_beta_dna_sequence_service import V2BetaDnaSequenceService
from benchling_sdk.services.v2.beta.v2_beta_entity_service import V2BetaEntityService
from benchling_sdk.services.v2.beta.v2_beta_entry_service import V2BetaEntryService
from benchling_sdk.services.v2.beta.v2_beta_rna_oligo_service import V2BetaRnaOligoService
from benchling_sdk.services.v2.beta.v2_beta_worklist_service import V2BetaWorklistService


class V2BetaService(BaseService):
    """
    V2-beta.

    Beta endpoints have different stability guidelines than other stable endpoints.

    See https://benchling.com/api/v2-beta/reference
    """

    _aa_sequence_service: V2BetaAaSequenceService
    _app_service: V2BetaAppService
    _custom_entity_service: V2BetaCustomEntityService
    _dna_oligo_service: V2BetaDnaOligoService
    _dna_sequence_service: V2BetaDnaSequenceService
    _entity_service: V2BetaEntityService
    _entry_service: V2BetaEntryService
    _rna_oligo_service: V2BetaRnaOligoService
    _worklist_service: V2BetaWorklistService

    def __init__(self, client: Client, retry_strategy: RetryStrategy = RetryStrategy()):
        """
        Initialize a v2-beta service.

        :param client: Underlying generated Client.
        :param retry_strategy: Retry strategy for failed HTTP calls
        """
        super().__init__(client, retry_strategy)
        beta_client = v2_beta_client(self.client)
        self._aa_sequence_service = V2BetaAaSequenceService(beta_client, retry_strategy)
        self._app_service = V2BetaAppService(beta_client, retry_strategy)
        self._custom_entity_service = V2BetaCustomEntityService(beta_client, retry_strategy)
        self._dna_sequence_service = V2BetaDnaSequenceService(beta_client, retry_strategy)
        self._dna_oligo_service = V2BetaDnaOligoService(beta_client, retry_strategy)
        self._entity_service = V2BetaEntityService(beta_client, retry_strategy)
        self._entry_service = V2BetaEntryService(beta_client, retry_strategy)
        self._rna_oligo_service = V2BetaRnaOligoService(beta_client, retry_strategy)
        self._worklist_service = V2BetaWorklistService(beta_client, retry_strategy)

    @property
    def aa_sequences(self) -> V2BetaAaSequenceService:
        """
        V2-Beta AA Sequences.

        AA Sequences are the working units of cells that make everything run (they help make structures, catalyze
        reactions and allow for signaling - a kind of internal cell communication). On Benchling, these are comprised
        of a string of amino acids and collections of other attributes, such as annotations.

        See https://benchling.com/api/v2-beta/reference#/AA%20Sequences
        """
        return self._aa_sequence_service

    @property
    def apps(self) -> V2BetaAppService:
        """
        V2-Beta Apps.

        Create and manage Apps on your tenant.

        https://benchling.com/api/v2-beta/reference?stability=not-available#/Apps
        """
        return self._app_service

    @property
    def custom_entities(self) -> V2BetaCustomEntityService:
        """
        V2-Beta Custom Entities.

        Benchling supports custom entities for biological entities that are neither sequences or proteins. Custom
        entities must have an entity schema set and can have both schema fields and custom fields.

        See https://benchling.com/api/v2-beta/reference#/Custom%20Entities
        """
        return self._custom_entity_service

    @property
    def dna_oligos(self) -> V2BetaDnaOligoService:
        """
        V2-Beta DNA Oligos.

        DNA Oligos are short linear DNA sequences that can be attached as primers to full DNA sequences. Just like other
        entities, they support schemas, tags, and aliases.

        See https://benchling.com/api/v2-beta/reference#/DNA%20Oligos
        """
        return self._dna_oligo_service

    @property
    def dna_sequences(self) -> V2BetaDnaSequenceService:
        """
        V2-Beta DNA Sequences.

        DNA sequences are the bread and butter of the Benchling Molecular Biology suite. On Benchling, these are
        comprised of a string of nucleotides and collections of other attributes, such as annotations and primers.

        See https://benchling.com/api/v2-beta/reference#/DNA%20Sequences
        """
        return self._dna_sequence_service

    @property
    def entities(self) -> V2BetaEntityService:
        """
        V2-Beta Entities.

        Entities include DNA and AA sequences, oligos, molecules, custom entities, and
        other biological objects in Benchling. Entities support schemas, tags, and aliases,
        and can be registered.

        See https://benchling.com/api/v2-beta/reference#/Entities
        """
        return self._entity_service

    @property
    def entries(self) -> V2BetaEntryService:
        """
        V2-Beta Entries.

        Entries are rich text documents that allow you to capture all of your experimental data in one place.

        https://benchling.com/api/v2-beta/reference#/Entries
        """
        return self._entry_service

    @property
    def rna_oligos(self) -> V2BetaRnaOligoService:
        """
        V2-Beta RNA Oligos.

        RNA Oligos are short linear RNA sequences that can be attached as primers to full DNA sequences. Just like other
        entities, they support schemas, tags, and aliases.

        See https://benchling.com/api/v2-beta/reference#/RNA%20Oligos
        """
        return self._rna_oligo_service

    @property
    def worklists(self) -> V2BetaWorklistService:
        """
        V2-Beta Worklists.

        Worklists are a convenient way to organize items for bulk actions, and are complementary to folders and
        projects.

        See https://benchling.com/api/v2-beta/reference#/Worklists
        """
        return self._worklist_service



from functools import wraps
import inspect
import logging
from typing import Any, Callable, cast, TypeVar

from benchling_sdk.helpers.logging_helpers import sdk_logger
from benchling_sdk.helpers.retry_helpers import retry_method

F = TypeVar("F", bound=Callable[..., Any])


def api_method(f: F) -> F:
    """Mark the decorated method as making one or more API calls to Benchling."""

    @wraps(f)
    @retry_method
    def wrapper(*args, **kwargs):
        _log_function(f, args, kwargs)
        return f(*args, **kwargs)

    return cast(F, wrapper)


def _log_function(func, args, kwargs) -> None:
    func_module = inspect.getmodule(func)
    module_name = "" if func_module is None else func_module.__name__
    sdk_logger.info("Calling %s in %s", func.__name__, module_name)
    if sdk_logger.isEnabledFor(logging.DEBUG):
        formatted_args = list(args)
        # The first value is a pointer to the service and object hash like
        # <benchling_sdk.services.schema.location_schema_service.LocationSchemaService object at 0x109c9a8e0>
        formatted_args.pop(0)
        sdk_logger.debug(
            "Calling %s in %s with arguments: %s keyword arguments: %s",
            func.__name__,  # Example: location_schemas_page
            module_name,  # Example: benchling_sdk.services.schema.location_schema_service
            formatted_args,  # Example: ["arg1", "arg2"]
            kwargs,  # Example: {"next_token": "", "page_size": 50}
        )



from io import BytesIO
from pathlib import Path

from benchling_api_client.v2.alpha.models.benchling_app_manifest import BenchlingAppManifest
import yaml


def manifest_to_bytes(manifest: BenchlingAppManifest, encoding: str = "utf-8") -> BytesIO:
    """Write a modeled Benchling App manifest to BytesIO of YAML."""
    manifest_dict = manifest.to_dict()
    yaml_format = yaml.safe_dump(manifest_dict, encoding=encoding, allow_unicode=True, sort_keys=False)
    return BytesIO(yaml_format)


def manifest_from_bytes(manifest: BytesIO) -> BenchlingAppManifest:
    """Read a modeled Benchling App manifest from BytesIO."""
    yaml_format = yaml.safe_load(manifest)
    return BenchlingAppManifest.from_dict(yaml_format)


def manifest_from_file(file_path: Path) -> BenchlingAppManifest:
    """Read a modeled Benchling App manifest from a file."""
    with open(file_path, "rb") as file:
        return manifest_from_bytes(BytesIO(file.read()))



from dataclasses import dataclass
from typing import Any, Dict, Iterable, Optional, Type, TypeVar, Union

from benchling_api_client.v2.types import Unset, UNSET
from dataclasses_json import DataClassJsonMixin

from benchling_sdk.models import CustomFields, Field, Fields, SchemaFieldsQueryParam

D = TypeVar("D", bound="DeserializableModel")
T = TypeVar("T")


@dataclass
class SerializableModel(DataClassJsonMixin):
    """
    Provide an interface for serialization of a custom model.

    For serializing models when using raw API calls (e.g., Benchling.api).
    Override serialize() for customized behavior.
    """

    # Not named to_dict() to avoid conflicts with DataClassJsonMixin
    def serialize(self) -> Dict[str, Any]:
        """Serialize the class attributes into a dict."""
        all_values = super().to_dict()
        # Filter out any Unset values
        return {key: value for key, value in all_values.items() if not isinstance(value, Unset)}


@dataclass
class DeserializableModel(DataClassJsonMixin):
    """Provide an interface for deserialization to a custom model.

    For deserializing models when using raw API calls (e.g., Benchling.api)
    Override deserialize() for customized behavior.
    """

    # Not named from_dict() to avoid conflicts with DataClassJsonMixin
    @classmethod
    def deserialize(cls: Type[D], source_dict: Dict[str, Any]) -> D:
        """Deserialize the input dictionary into the model."""
        return cls.from_dict(source_dict)


@dataclass
class DeserializableModelNoContent(DeserializableModel):
    """Provide an interface deserialization to a custom model when the response may be empty."""

    pass


def optional_array_query_param(inputs: Optional[Iterable[str]]) -> Optional[str]:
    """Collapse an Iterable to a comma-separated string if present.

    Add leading and trailing quotes if the item contains "," and not quoted
    """
    return array_query_param(inputs) if inputs is not None else None


def array_query_param(inputs: Iterable[str]) -> str:
    """Collapse an Iterable to a comma-separated string.

    Add leading and trailing quotes if the item contains "," and not quoted
    """
    quoted_inputs = ['"' + item + '"' if "," in item and item[0] != '"' else item for item in inputs]
    return ",".join(quoted_inputs)


def fields(source: Dict[str, Dict[str, Any]]) -> Fields:
    """Marshal a dictionary into a Fields object."""
    deserialized_fields = Fields()
    for field_name, field_dict in source.items():
        # BNCH-22297: Workaround to deal with the fact that `Field` is not correctly deserialized.
        deserialized_fields[field_name] = Field(value=field_dict["value"])
    return deserialized_fields


def custom_fields(source: Dict[str, Any]) -> CustomFields:
    """Marshal a dictionary into a CustomFields object."""
    return CustomFields.from_dict(source)


def unset_as_none(source: Union[Unset, None, T]) -> Optional[T]:
    """Given a value that may be UNSET, produces an Optional[] where UNSET will be treated as None."""
    if source is UNSET:
        return None
    return source  # type: ignore


def schema_fields_query_param(schema_fields: Optional[Dict[str, Any]]) -> Optional[SchemaFieldsQueryParam]:
    """Prefix keys with 'schemaField' for querying fields in list endpoints."""
    return (
        SchemaFieldsQueryParam.from_dict(
            {f"schemaField.{field}": value for field, value in schema_fields.items()}
        )
        if schema_fields
        else None
    )


def none_as_unset(val: Optional[T]) -> Union[T, Unset]:
    """Convert a value to UNSET if None safely."""
    return UNSET if val is None else val



from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.folders import (
    archive_folders,
    create_folder,
    get_folder,
    list_folders,
    unarchive_folders,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import (
    Folder,
    FolderCreate,
    FoldersArchivalChange,
    FoldersArchive,
    FoldersArchiveReason,
    FoldersPaginatedList,
    FoldersUnarchive,
    ListFoldersSection,
    ListFoldersSort,
)
from benchling_sdk.services.v2.base_service import BaseService


class FolderService(BaseService):
    """
    Folders.

    Manage folder objects.

    See https://benchling.com/api/reference#/Folders
    """

    @api_method
    def get_by_id(self, folder_id: str) -> Folder:
        """
        Get a folder by ID.

        See https://benchling.com/api/reference#/Folders/getFolder
        """
        response = get_folder.sync_detailed(client=self.client, folder_id=folder_id)
        return model_from_detailed(response)

    @api_method
    def _folders_page(
        self,
        *,
        sort: Optional[ListFoldersSort] = ListFoldersSort.NAME,
        archive_reason: Optional[str] = None,
        name_includes: Optional[str] = None,
        parent_folder_id: Optional[str] = None,
        project_id: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        section: Optional[ListFoldersSection] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[FoldersPaginatedList]:
        response = list_folders.sync_detailed(
            client=self.client,
            sort=none_as_unset(sort),
            archive_reason=none_as_unset(archive_reason),
            name_includes=none_as_unset(name_includes),
            parent_folder_id=none_as_unset(parent_folder_id),
            project_id=none_as_unset(project_id),
            ids=none_as_unset(optional_array_query_param(ids)),
            name=none_as_unset(name),
            section=none_as_unset(section),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response

    def list(
        self,
        *,
        sort: Optional[ListFoldersSort] = None,
        archive_reason: Optional[str] = None,
        name_includes: Optional[str] = None,
        parent_folder_id: Optional[str] = None,
        project_id: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        section: Optional[ListFoldersSection] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[Folder]:
        """
        List folders.

        See https://benchling.com/api/reference#/Folders/listFolders
        """

        def api_call(next_token: NextToken) -> Response[FoldersPaginatedList]:
            return self._folders_page(
                sort=sort,
                archive_reason=archive_reason,
                name_includes=name_includes,
                parent_folder_id=parent_folder_id,
                project_id=project_id,
                ids=ids,
                name=name,
                section=section,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: FoldersPaginatedList) -> Optional[List[Folder]]:
            return body.folders

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, folder: FolderCreate) -> Folder:
        """
        Create folder.

        See https://benchling.com/api/reference#/Folders/createFolder
        """
        response = create_folder.sync_detailed(client=self.client, json_body=folder)
        return model_from_detailed(response)

    @api_method
    def archive(self, folder_ids: Iterable[str], reason: FoldersArchiveReason) -> FoldersArchivalChange:
        """
        Archive folders.

        See https://benchling.com/api/reference#/Folders/archiveFolders
        """
        archive_request = FoldersArchive(folder_ids=list(folder_ids), reason=reason)
        response = archive_folders.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, folder_ids: Iterable[str]) -> FoldersArchivalChange:
        """
        Unarchive folders.

        See https://benchling.com/api/reference#/Folders/unarchiveFolders
        """
        unarchive_request = FoldersUnarchive(folder_ids=list(folder_ids))
        response = unarchive_folders.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)



from enum import Enum
import logging
from typing import Iterable, Optional
import warnings


class StabilityWarning(Warning):
    """Warning for users that a preview feature is being used."""

    pass


class StabilityLevel(Enum):
    """Stability levels for discrete guidelines."""

    ALPHA = "alpha"
    BETA = "beta"


def default_logger(name: str = "benchling_sdk") -> logging.Logger:
    """Construct the default logger for the SDK."""
    logger = logging.getLogger(name)
    logger.addHandler(logging.NullHandler())
    return logger


sdk_logger = default_logger()


def log_deprecation(source_name: str, suggestion: str) -> None:
    """Log a standardized warning for use of deprecated functionality."""
    warnings.warn(
        message=f"{source_name} is deprecated. Please use {suggestion}",
        category=DeprecationWarning,
        # Indicate the line in the caller's code, not this method or the SDK method referencing it
        stacklevel=3,
    )


def log_stability_warning(level: StabilityLevel, package: str = "") -> None:
    """Log a standardized warning for using preview functionality."""
    prefix = "This package"
    if package:
        prefix = f"{prefix} {package}"
    warnings.warn(
        message=f"{prefix} is in {level.value} and not recommended for production use. "
        f"See our stability guidelines for more information: https://docs.benchling.com/docs/stability",
        category=StabilityWarning,
        # Indicate the line in the caller's code, not this method or the SDK method referencing it
        stacklevel=3,
    )


# This doesn't really belong in logging helpers, but we're not moving it for compatibility
# just in case someone else is importing and using it


def check_for_csv_bug_fix(variable: str, value: Optional[Iterable[str]]):
    """Specific error for a fixed bug that might cause breaks. See BNCH-24087."""
    if value:
        value_list = list(value)
        if len(value_list) == 1 and "," in value_list[0]:
            raise ValueError(f"Items in {variable} must not contain ','")



from typing import Any, Dict, Iterable, List, Optional

from benchling_api_client.v2.stable.api.aa_sequences import (
    archive_aa_sequences,
    auto_annotate_aa_sequences,
    bulk_create_aa_sequences,
    bulk_get_aa_sequences,
    bulk_update_aa_sequences,
    create_aa_sequence,
    get_aa_sequence,
    list_aa_sequences,
    unarchive_aa_sequences,
    update_aa_sequence,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import check_for_csv_bug_fix
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.models import (
    AaSequence,
    AaSequenceBulkCreate,
    AaSequenceBulkUpdate,
    AaSequenceCreate,
    AaSequencesArchivalChange,
    AaSequencesArchive,
    AaSequencesBulkCreateRequest,
    AaSequencesBulkUpdateRequest,
    AaSequencesPaginatedList,
    AaSequencesUnarchive,
    AaSequenceUpdate,
    AsyncTaskLink,
    AutoAnnotateAaSequences,
    EntityArchiveReason,
    ListAASequencesSort,
)
from benchling_sdk.services.v2.base_service import BaseService


class AaSequenceService(BaseService):
    """
    AA Sequences.

    AA Sequences are the working units of cells that make everything run (they help make structures, catalyze
    reactions and allow for signaling - a kind of internal cell communication). On Benchling, these are comprised
    of a string of amino acids and collections of other attributes, such as annotations.
    See https://benchling.com/api/reference#/AA%20Sequences
    """

    @api_method
    def get_by_id(self, aa_sequence_id: str) -> AaSequence:
        """
        Get an AA Sequence.

        See https://benchling.com/api/reference#/AA%20Sequences/getAASequence
        """
        response = get_aa_sequence.sync_detailed(client=self.client, aa_sequence_id=aa_sequence_id)
        return model_from_detailed(response)

    @api_method
    def _aa_sequences_page(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        amino_acids: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        sort: Optional[ListAASequencesSort] = None,
        page_size: Optional[int] = None,
        next_token: NextToken = None,
        author_idsany_of: Optional[Iterable[str]] = None,
    ) -> Response[AaSequencesPaginatedList]:
        response = list_aa_sequences.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            amino_acids=none_as_unset(amino_acids),
            folder_id=none_as_unset(folder_id),
            mentioned_in=none_as_unset(optional_array_query_param(mentioned_in)),
            project_id=none_as_unset(project_id),
            registry_id=none_as_unset(registry_id),
            schema_id=none_as_unset(schema_id),
            archive_reason=none_as_unset(archive_reason),
            mentions=none_as_unset(optional_array_query_param(mentions)),
            ids=none_as_unset(optional_array_query_param(ids)),
            entity_registry_idsany_of=none_as_unset(optional_array_query_param(entity_registry_ids_any_of)),
            name_includes=none_as_unset(name_includes),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            sort=none_as_unset(sort),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            author_idsany_of=none_as_unset(optional_array_query_param(author_idsany_of)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        amino_acids: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        sort: Optional[ListAASequencesSort] = None,
        page_size: Optional[int] = None,
        author_idsany_of: Optional[Iterable[str]] = None,
    ) -> PageIterator[AaSequence]:
        """
        List AA Sequences.

        See https://benchling.com/api/reference#/AA%20Sequences/listAASequences
        """
        check_for_csv_bug_fix("mentioned_in", mentioned_in)
        check_for_csv_bug_fix("mentions", mentions)

        def api_call(next_token: NextToken) -> Response[AaSequencesPaginatedList]:
            return self._aa_sequences_page(
                modified_at=modified_at,
                name=name,
                amino_acids=amino_acids,
                folder_id=folder_id,
                mentioned_in=mentioned_in,
                project_id=project_id,
                registry_id=registry_id,
                schema_id=schema_id,
                archive_reason=archive_reason,
                mentions=mentions,
                ids=ids,
                entity_registry_ids_any_of=entity_registry_ids_any_of,
                name_includes=name_includes,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                schema_fields=schema_fields,
                sort=sort,
                page_size=page_size,
                next_token=next_token,
                creator_ids=creator_ids,
                author_idsany_of=author_idsany_of,
            )

        def results_extractor(body: AaSequencesPaginatedList) -> Optional[List[AaSequence]]:
            return body.aa_sequences

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, aa_sequence: AaSequenceCreate) -> AaSequence:
        """
        Create an AA Sequence.

        See https://benchling.com/api/reference#/AA%20Sequences/createAASequence
        """
        response = create_aa_sequence.sync_detailed(client=self.client, json_body=aa_sequence)
        return model_from_detailed(response)

    @api_method
    def update(self, aa_sequence_id: str, aa_sequence: AaSequenceUpdate) -> AaSequence:
        """
        Update an AA Sequence.

        See https://benchling.com/api/reference#/AA%20Sequences/updateAASequence
        """
        response = update_aa_sequence.sync_detailed(
            client=self.client, aa_sequence_id=aa_sequence_id, json_body=aa_sequence
        )
        return model_from_detailed(response)

    @api_method
    def archive(
        self, aa_sequence_ids: Iterable[str], reason: EntityArchiveReason
    ) -> AaSequencesArchivalChange:
        """
        Archive an AA Sequence.

        See https://benchling.com/api/reference#/AA%20Sequences/archiveAASequences
        """
        archive_request = AaSequencesArchive(reason=reason, aa_sequence_ids=list(aa_sequence_ids))
        response = archive_aa_sequences.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, aa_sequence_ids: Iterable[str]) -> AaSequencesArchivalChange:
        """
        Unarchive an AA Sequence.

        See https://benchling.com/api/reference#/AA%20Sequences/unarchiveAASequences
        """
        unarchive_request = AaSequencesUnarchive(aa_sequence_ids=list(aa_sequence_ids))
        response = unarchive_aa_sequences.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def bulk_get(self, aa_sequence_ids: Iterable[str]) -> Optional[List[AaSequence]]:
        """
        Bulk get AA sequences by ID.

        See https://benchling.com/api/reference#/AA%20Sequences/bulkGetAASequences
        """
        aa_sequence_id_string = ",".join(aa_sequence_ids)
        response = bulk_get_aa_sequences.sync_detailed(
            client=self.client, aa_sequence_ids=aa_sequence_id_string
        )
        aa_sequences_results = model_from_detailed(response)
        return aa_sequences_results.aa_sequences

    @api_method
    def bulk_create(self, aa_sequences: Iterable[AaSequenceBulkCreate]) -> AsyncTaskLink:
        """
        Bulk create AA sequences.

        See https://benchling.com/api/reference#/AA%20Sequences/bulkCreateAASequences
        """
        body = AaSequencesBulkCreateRequest(list(aa_sequences))
        response = bulk_create_aa_sequences.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_update(self, aa_sequences: Iterable[AaSequenceBulkUpdate]) -> AsyncTaskLink:
        """
        Bulk update AA sequences.

        See https://benchling.com/api/reference#/AA%20Sequences/bulkUpdateAASequences
        """
        body = AaSequencesBulkUpdateRequest(list(aa_sequences))
        response = bulk_update_aa_sequences.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def auto_annotate(self, auto_annotate: AutoAnnotateAaSequences) -> AsyncTaskLink:
        """
        Auto-annotate AA sequences with matching features from specified Feature Libraries.

        See https://benchling.com/api/reference#/AA%20Sequences/autoAnnotateAaSequences
        """
        response = auto_annotate_aa_sequences.sync_detailed(client=self.client, json_body=auto_annotate)
        return model_from_detailed(response)



from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status


def model_from_detailed(response: Response):
    """Deserialize a response into a model."""
    raise_for_status(response)
    return response.parsed



from typing import Any, Dict, Optional, Type, TypeVar
import urllib
from urllib.parse import ParseResult

from benchling_api_client.v2.stable.client import Client
from benchling_api_client.v2.types import Response
import httpx
from httpx._types import RequestData, RequestFiles

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.serialization_helpers import (
    DeserializableModel,
    DeserializableModelNoContent,
    SerializableModel,
)
from benchling_sdk.services.v2.base_service import BaseService

D = TypeVar("D", bound=DeserializableModel)
S = TypeVar("S", bound=SerializableModel)


class ApiService(BaseService):
    """
    A service for making raw API calls to Benchling using an authenticated client.

    This is generally useful for using the SDK to call API endpoints which may not be supported in the SDK yet.
    """

    def _get(self, url: str, additional_headers: Optional[Dict[str, Any]] = None) -> httpx.Response:
        """Return a raw HTTPX response. Does not assume JSON. Does not automatically retry errors."""
        args = _get_kwargs(self.client, url, additional_headers)
        return self.client.httpx_client.get(**args)

    @api_method
    def get_response(
        self, url: str, additional_headers: Optional[Dict[str, Any]] = None
    ) -> Response[Dict[str, Any]]:
        """
        Call the API with HTTP GET. Return a Response with JSON deserialized as Dict[str, Any].

        :param url: A relative URL path excluding host and scheme (e.g., /api/v2-alpha/some-endpoint) :param
        additional_headers: A dict of key/values to add to the request as HTTP headers :return: The raw HTTP Response
        :rtype: Response[Dict[str, Any]]
        """
        httpx_response = self._get(url, additional_headers)
        response = build_json_response(response=httpx_response)
        raise_for_status(response)
        return response

    @api_method
    def get_modeled(
        self, url: str, target_type: Type[D], additional_headers: Optional[Dict[str, Any]] = None
    ) -> D:
        """
        Return a deserialized model as specified by target_type.

        Automatically retries errors as configured in RetryStrategy.

        :param url: A relative URL path excluding host and scheme (e.g., /api/v2-alpha/some-endpoint) :param
        target_type: A class type of DeserializableModel to deserialize from the response JSON :param
        additional_headers: A dict of key/values to add to the request as HTTP headers :return: The deserialized
        modeled response :rtype: D
        """
        response = self.get_response(url, additional_headers)
        optional_parsed = response.parsed
        assert (
            optional_parsed is not None
        ), f"The response from {url} was empty, could not parse {target_type}"
        return target_type.deserialize(optional_parsed)

    def _post(
        self,
        url: str,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Dict[str, Any]] = None,
        additional_headers: Optional[Dict[str, Any]] = None,
    ) -> httpx.Response:
        """Return a raw HTTPX response. Does not assume JSON. Does not automatically retry errors."""
        args = _body_kwargs(
            self.client, url, data=data, files=files, json=json, additional_headers=additional_headers
        )
        return self.client.httpx_client.post(**args)

    @api_method
    def post_response(
        self,
        url: str,
        body: Optional[Dict[str, Any]] = None,
        additional_headers: Optional[Dict[str, Any]] = None,
    ) -> Response[Dict[str, Any]]:
        """
        Call the API with HTTP POST. Return a Response with JSON deserialized as Dict[str, Any].

        :param url: A relative URL path excluding host and scheme (e.g., /api/v2-alpha/some-endpoint) :param body: Body
        of the POST request, which will be serialized to JSON :param additional_headers: A dict of key/values to add to
        the request as HTTP headers :return: The raw HTTP Response :rtype: Response[Dict[str, Any]]
        """
        httpx_response = self._post(url, json=body, additional_headers=additional_headers)
        response = build_json_response(response=httpx_response)
        raise_for_status(response)
        return response

    @api_method
    def post_modeled(
        self,
        url: str,
        target_type: Type[D],
        body: Optional[S] = None,
        additional_headers: Optional[Dict[str, Any]] = None,
    ) -> Optional[D]:
        """
        Return a deserialized model as specified by target_type.

        Automatically retries errors as configured in RetryStrategy.

        :param url: A relative URL path excluding host and scheme (e.g., /api/v2-alpha/some-endpoint) :param
        target_type: A class type of DeserializableModel to deserialize from the response JSON :param body: Body of the
        POST request as a class implementing SerializableModel. Will be serialized to JSON :param additional_headers: A
        dict of key/values to add to the request as HTTP headers :return: The deserialized modeled response :rtype: D
        """
        serialized_body = body.serialize() if body else None
        response = self.post_response(url, serialized_body, additional_headers)
        optional_parsed = response.parsed
        return target_type.deserialize(optional_parsed) if optional_parsed else None

    def _patch(
        self,
        url: str,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Dict[str, Any]] = None,
        additional_headers: Optional[Dict[str, Any]] = None,
    ) -> httpx.Response:
        """Return a raw HTTPX response. Does not assume JSON. Does not automatically retry errors."""
        args = _body_kwargs(
            self.client, url, data=data, files=files, json=json, additional_headers=additional_headers
        )
        return self.client.httpx_client.patch(**args)

    @api_method
    def patch_response(
        self,
        url: str,
        body: Optional[Dict[str, Any]] = None,
        additional_headers: Optional[Dict[str, Any]] = None,
    ) -> Response[Dict[str, Any]]:
        """
        Call the API with HTTP PATCH. Returns a Response with JSON deserialized as Dict[str, Any].

        :param url: A relative URL path excluding host and scheme (e.g., /api/v2-alpha/some-endpoint) :param body: Body
        of the PATCH request, which will be serialized to JSON :param additional_headers: A dict of key/values to add
        to the request as HTTP headers :return: The raw HTTP Response :rtype: Response[Dict[str, Any]]
        """
        httpx_response = self._patch(url, json=body, additional_headers=additional_headers)
        response = build_json_response(response=httpx_response)
        raise_for_status(response)
        return response

    @api_method
    def patch_modeled(
        self,
        url: str,
        target_type: Type[D],
        body: Optional[S] = None,
        additional_headers: Optional[Dict[str, Any]] = None,
    ) -> Optional[D]:
        """
        Return a deserialized model as specified by target_type.

        Automatically retries errors as configured in RetryStrategy.

        :param url: A relative URL path excluding host and scheme (e.g., /api/v2-alpha/some-endpoint) :param
        target_type: A class type of DeserializableModel to deserialize from the response JSON :param body: Body of the
        PATCH request as a class implementing SerializableModel. Will be serialized to JSON :param additional_headers:
        A dict of key/values to add to the request as HTTP headers :return: The deserialized modeled response :rtype: D
        """
        serialized_body = body.serialize() if body else None
        response = self.patch_response(url, serialized_body, additional_headers)
        optional_parsed = response.parsed
        return target_type.deserialize(optional_parsed) if optional_parsed else None

    def _delete(
        self,
        url: str,
        additional_headers: Optional[Dict[str, Any]] = None,
    ) -> httpx.Response:
        """
        Return a raw HTTPX response. Does not assume JSON. Does not automatically retry errors.

        :param url: A relative URL path excluding host and scheme (e.g., /api/v2-alpha/some-endpoint) :param
        additional_headers: A dict of key/values to add to the request as HTTP headers
        """
        args = _get_kwargs(self.client, url, additional_headers=additional_headers)
        return self.client.httpx_client.delete(**args)

    @api_method
    def delete_response(
        self,
        url: str,
        additional_headers: Optional[Dict[str, Any]] = None,
    ) -> Response[Dict[str, Any]]:
        """Call the API with HTTP DELETE. Returns a Response with JSON deserialized as Dict[str, Any]."""
        httpx_response = self._delete(url, additional_headers=additional_headers)
        response = build_json_response(response=httpx_response)
        raise_for_status(response)
        return response

    @api_method
    def delete_modeled(
        self,
        url: str,
        target_type: Type[D] = DeserializableModelNoContent,  # type: ignore
        additional_headers: Optional[Dict[str, Any]] = None,
    ) -> Optional[D]:
        """
        Return a deserialized model as specified by target_type.

        Automatically retries errors as configured in RetryStrategy.

        :param url: A relative URL path excluding host and scheme (e.g., /api/v2-alpha/some-endpoint) :param
        target_type: A class type of DeserializableModel to deserialize from the response JSON :param
        additional_headers: A dict of key/values to add to the request as HTTP headers :return: The deserialized
        modeled response :rtype: D
        """
        response = self.delete_response(url, additional_headers)
        optional_parsed = response.parsed
        return target_type.deserialize(optional_parsed) if optional_parsed else None


def _get_kwargs(client: Client, url: str, additional_headers: Optional[Dict[str, Any]]) -> Dict[str, Any]:
    """Prepare arguments for HTTPX get calls."""
    full_url = _build_url(client, url)
    headers = client.get_headers()
    if additional_headers:
        headers.update(additional_headers)
    return {
        "url": full_url,
        "headers": headers,
        "cookies": client.get_cookies(),
        "timeout": client.get_timeout(),
    }


def _body_kwargs(
    client: Client,
    url: str,
    data: Optional[RequestData],
    files: Optional[RequestFiles],
    json: Optional[Dict[str, Any]],
    additional_headers: Optional[Dict[str, Any]],
) -> Dict[str, Any]:
    """Prepare arguments for HTTPX post, put, patch calls with bodies."""
    params = _get_kwargs(client, url, additional_headers)
    body_params: Dict[str, Any] = dict()
    if json:
        body_params["json"] = json
    if data:
        body_params["data"] = data
    if files:
        body_params["files"] = files
    params.update(body_params)
    return params


def _build_url(client: Client, url: str) -> str:
    """
    Strip the Client's base URL to the schema and server, then append a fully qualified URL with the new path.

    For example called with 'new/path' and a client configured like: client.base_url = 'http://benchling.com/some/path'

    The result will be: 'http://benchling.com/new/path'
    """
    parse_result: ParseResult = urllib.parse.urlparse(client.base_url)
    full_host = parse_result.hostname
    if parse_result.port:
        full_host = f"{full_host}:{parse_result.port}"
    # Avoid any repeated slashes
    path = f"{full_host}/{url}".replace("//", "/")
    return f"{parse_result.scheme}://{path}"


def build_json_response(response: httpx.Response) -> Response[Dict[str, Any]]:
    """
    Create a Response object which corresponds to the usual code generated response.

    This assumes a JSON return payload but will structure a little more than the raw HTTPX response.
    """
    parsed_response = None if response.is_error or not response.text else response.json()
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=parsed_response,
    )



from benchling_api_client.v2.stable.client import Client

from benchling_sdk.helpers.logging_helpers import log_stability_warning, StabilityLevel
from benchling_sdk.helpers.retry_helpers import RetryStrategy
from benchling_sdk.services.v2.base_service import BaseService
from benchling_sdk.services.v2.v2_alpha_service import V2AlphaService
from benchling_sdk.services.v2.v2_beta_service import V2BetaService
from benchling_sdk.services.v2.v2_stable_service import V2StableService


class V2Service(BaseService):
    """
    V2.

    Namespace containing support for the V2 Benchling API.
    """

    _v2_alpha_service: V2AlphaService
    _v2_beta_service: V2BetaService
    _v2_stable_service: V2StableService

    def __init__(self, client: Client, retry_strategy: RetryStrategy = RetryStrategy()):
        """
        Initialize a service.

        :param client: Underlying generated Client.
        :param retry_strategy: Retry strategy for failed HTTP calls
        """
        super().__init__(client, retry_strategy)
        self._v2_stable_service = V2StableService(client, retry_strategy)
        self._v2_alpha_service = V2AlphaService(client, retry_strategy)
        self._v2_beta_service = V2BetaService(client, retry_strategy)

    @property
    def alpha(self) -> V2AlphaService:
        """
        V2-alpha.

        Alpha endpoints have different stability guidelines than other stable endpoints.

        See https://benchling.com/api/v2-alpha/reference
        """
        log_stability_warning(StabilityLevel.ALPHA, package="v2.alpha")
        return self._v2_alpha_service

    @property
    def beta(self) -> V2BetaService:
        """
        V2-beta.

        Beta endpoints have different stability guidelines than other stable endpoints.

        See https://benchling.com/api/v2-beta/reference
        """
        log_stability_warning(StabilityLevel.BETA, package="v2.beta")
        return self._v2_beta_service

    @property
    def stable(self) -> V2StableService:
        """
        Stable.

        See https://docs.benchling.com/docs/stability#are-breaking-changes-made-to-the-api-or-events
        """
        return self._v2_stable_service



from benchling_api_client.v2.stable.api.warehouse import create_warehouse_credentials

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.models import WarehouseCredentials, WarehouseCredentialsCreate
from benchling_sdk.services.v2.base_service import BaseService


class WarehouseService(BaseService):
    """
    Warehouse.

    Manage warehouse credentials.

    See https://benchling.com/api/reference#/Warehouse
    """

    @api_method
    def create_credentials(self, credentials: WarehouseCredentialsCreate) -> WarehouseCredentials:
        """
        Create Benchling Warehouse credentials.

        Allows for programmatically generating credentials to connect to the Benchling warehouse.
        You must have a warehouse configured to access this endpoint.

        The credentials will authenticate as the same user calling the API. Note that `expires_in` is required -
        only temporary credentials are currently allowed.

        See https://benchling.com/api/reference#/Warehouse/createWarehouseCredentials
        """
        response = create_warehouse_credentials.sync_detailed(client=self.client, json_body=credentials)
        return model_from_detailed(response)



from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.dropdowns import (
    archive_dropdown_options,
    create_dropdown,
    get_dropdown,
    list_dropdowns,
    unarchive_dropdown_options,
    update_dropdown,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset
from benchling_sdk.models import (
    Dropdown,
    DropdownCreate,
    DropdownOptionsArchivalChange,
    DropdownOptionsArchive,
    DropdownOptionsArchiveReason,
    DropdownOptionsUnarchive,
    DropdownSummariesPaginatedList,
    DropdownSummary,
    DropdownUpdate,
)
from benchling_sdk.services.v2.base_service import BaseService


class DropdownService(BaseService):
    """
    Dropdowns.

    Dropdowns are registry-wide enums. Use dropdowns to standardize on spelling and naming conventions, especially
    for important metadata like resistance markers.

    See https://benchling.com/api/reference#/Dropdowns
    """

    @api_method
    def get_by_id(self, dropdown_id: str) -> Dropdown:
        """
        Get a dropdown.

        See https://benchling.com/api/reference#/Dropdowns/getDropdown
        """
        response = get_dropdown.sync_detailed(client=self.client, dropdown_id=dropdown_id)
        return model_from_detailed(response)

    @api_method
    def _dropdowns_page(
        self,
        *,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
        registry_id: Optional[str] = None,
    ) -> Response[DropdownSummariesPaginatedList]:
        response = list_dropdowns.sync_detailed(
            client=self.client,
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
            registry_id=none_as_unset(registry_id),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        *,
        page_size: Optional[int] = 50,
        registry_id: Optional[str] = None,
    ) -> PageIterator[DropdownSummary]:
        """
        List dropdowns.

        See https://benchling.com/api/reference#/Dropdowns/listDropdowns
        """

        def api_call(next_token: NextToken) -> Response[DropdownSummariesPaginatedList]:
            return self._dropdowns_page(
                next_token=next_token,
                page_size=page_size,
                registry_id=registry_id,
            )

        def results_extractor(body: DropdownSummariesPaginatedList) -> Optional[List[DropdownSummary]]:
            return body.dropdowns

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, dropdown: DropdownCreate) -> Dropdown:
        """
        Create a dropdown.

        See https://benchling.com/api/reference#/Dropdowns/createDropdown
        """
        response = create_dropdown.sync_detailed(client=self.client, json_body=dropdown)
        return model_from_detailed(response)

    @api_method
    def update(self, dropdown_id: str, dropdown: DropdownUpdate) -> Dropdown:
        """
        Update a dropdown.

        See https://benchling.com/api/reference#/Dropdowns/updateDropdown
        """
        response = update_dropdown.sync_detailed(
            client=self.client, dropdown_id=dropdown_id, json_body=dropdown
        )
        return model_from_detailed(response)

    @api_method
    def archive_options(
        self, dropdown_id: str, dropdown_option_ids: Iterable[str], reason: DropdownOptionsArchiveReason
    ) -> DropdownOptionsArchivalChange:
        """
        Archive dropdown options.

        See https://benchling.com/api/reference#/Dropdowns/archiveDropdownOptions
        """
        dropdown_options_archive = DropdownOptionsArchive(
            reason=reason, dropdown_option_ids=list(dropdown_option_ids)
        )
        response = archive_dropdown_options.sync_detailed(
            client=self.client, dropdown_id=dropdown_id, json_body=dropdown_options_archive
        )
        return model_from_detailed(response)

    @api_method
    def unarchive_options(
        self, dropdown_id: str, dropdown_option_ids: Iterable[str]
    ) -> DropdownOptionsArchivalChange:
        """
        Unarchive dropdown options.

        See https://benchling.com/api/reference#/Dropdowns/unarchiveDropdownOptions
        """
        dropdown_options_unarchive = DropdownOptionsUnarchive(dropdown_option_ids=list(dropdown_option_ids))
        response = unarchive_dropdown_options.sync_detailed(
            client=self.client, dropdown_id=dropdown_id, json_body=dropdown_options_unarchive
        )
        return model_from_detailed(response)



from benchling_api_client.v2.beta.api.entries import create_entry
from benchling_api_client.v2.beta.models.entry import Entry
from benchling_api_client.v2.beta.models.entry_create import EntryCreate

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaEntryService(BaseService):
    """
    V2-Beta Entries.

    Entries are rich text documents that allow you to capture all of your experimental data in one place.

    https://benchling.com/api/v2-beta/reference#/Entries
    """

    @api_method
    def create_entry(self, entry: EntryCreate) -> Entry:
        """
        Create a notebook entry against beta endpoint, supporting initialTables.

        See https://benchling.com/api/v2-beta/reference#/Entries/createEntry
        """
        response = create_entry.sync_detailed(
            client=self.client,
            json_body=entry,
        )
        return model_from_detailed(response)



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Generic, List, Optional, Tuple, Type, TypeVar, Union

from benchling_api_client.v2.alpha.extensions import UnknownType
from benchling_api_client.v2.alpha.models.app_config_item import AppConfigItem
from benchling_api_client.v2.alpha.models.boolean_app_config_item import BooleanAppConfigItem
from benchling_api_client.v2.alpha.models.date_app_config_item import DateAppConfigItem
from benchling_api_client.v2.alpha.models.datetime_app_config_item import DatetimeAppConfigItem
from benchling_api_client.v2.alpha.models.entity_schema_app_config_item import EntitySchemaAppConfigItem
from benchling_api_client.v2.alpha.models.field_app_config_item import FieldAppConfigItem
from benchling_api_client.v2.alpha.models.float_app_config_item import FloatAppConfigItem
from benchling_api_client.v2.alpha.models.generic_app_config_item import GenericAppConfigItem
from benchling_api_client.v2.alpha.models.integer_app_config_item import IntegerAppConfigItem
from benchling_api_client.v2.alpha.models.json_app_config_item import JsonAppConfigItem
from benchling_api_client.v2.alpha.models.secure_text_app_config_item import SecureTextAppConfigItem
from typing_extensions import Protocol

from benchling_sdk.apps.config.decryption_provider import BaseDecryptionProvider
from benchling_sdk.apps.config.scalars import (
    DEFAULT_SCALAR_DEFINITIONS,
    ScalarConfigItemType,
    ScalarDefinition,
    ScalarType,
)
from benchling_sdk.benchling import Benchling


class MissingDependencyError(Exception):
    """
    Missing dependency error.

    Indicates a dependency was missing from app config.
    For instance, no dependency with that name was in the list.
    """

    pass


class UnsupportedDependencyError(Exception):
    """
    Unsupported dependency error.

    The manifest and configuration specified a dependency which the SDK is incapable of handling yet.
    """

    pass


class MissingScalarDefinitionError(Exception):
    """
    Missing scalar definition error.

    The manifest and configuration specified a scalar type which the SDK does not know how to translate
    to Python values yet.
    """

    pass


ConfigItemPath = Tuple[str, ...]

# Everything from AppConfigItem except UnknownType
ConfigurationReference = Union[
    DateAppConfigItem,
    DatetimeAppConfigItem,
    JsonAppConfigItem,
    EntitySchemaAppConfigItem,
    FieldAppConfigItem,
    BooleanAppConfigItem,
    IntegerAppConfigItem,
    FloatAppConfigItem,
    GenericAppConfigItem,
    SecureTextAppConfigItem,
]

D = TypeVar("D", bound="BaseDependencies")


class ConfigProvider(Protocol):
    """
    Config provider.

    Provides a BenchlingAppConfiguration.
    """

    def config(self) -> List[ConfigurationReference]:
        """Implement to provide a Benchling app configuration."""
        ...


class BenchlingConfigProvider(ConfigProvider):
    """
    Benchling Config provider.

    Provides a BenchlingAppConfiguration retrieved from Benchling's API.
    """

    _client: Benchling
    _app_id: str

    def __init__(self, client: Benchling, app_id: str):
        """
        Initialize Benchling Config Provider.

        :param client: A configured Benchling instance for making API calls.
        :param app_id: The app_id from which to retrieve configuration.
        """
        self._client = client
        self._app_id = app_id

    def config(self) -> List[ConfigurationReference]:
        """Provide a Benchling app configuration from Benchling's APIs."""
        app_pages = self._client.v2.alpha.apps.list_app_configuration_items(
            app_id=self._app_id, page_size=100
        )

        # Eager load all config items for now since we don't yet have a way of lazily querying by path
        all_config_pages = list(app_pages)
        # Punt on UnknownType for now as apps using manifests with new types could lead to unpredictable results
        all_config_items = [
            _supported_config_item(config_item) for page in all_config_pages for config_item in page
        ]

        return all_config_items


class StaticConfigProvider(ConfigProvider):
    """
    Static Config provider.

    Provides a BenchlingAppConfiguration from a static declaration. Useful for mocking or testing.
    """

    _configuration_items: List[ConfigurationReference]

    def __init__(self, configuration_items: List[ConfigurationReference]):
        """
        Initialize Static Config Provider.

        :param configuration_items: The configuration items to return.
        """
        self._configuration_items = configuration_items

    def config(self) -> List[ConfigurationReference]:
        """Provide Benchling app configuration items from a static reference."""
        return self._configuration_items


class DependencyLinkStore(object):
    """
    Dependency Link Store.

    Marshalls an app configuration from the configuration provider into an indexable structure.
    Only retrieves app configuration once unless its cache is invalidated.
    """

    _configuration_provider: ConfigProvider
    _configuration: Optional[List[ConfigurationReference]] = None
    _configuration_map: Optional[Dict[ConfigItemPath, ConfigurationReference]] = None

    def __init__(self, configuration_provider: ConfigProvider):
        """
        Initialize Dependency Link Store.

        :param configuration_provider: A ConfigProvider that will be invoked to provide the
        underlying config from which to organize dependency links.
        """
        self._configuration_provider = configuration_provider

    @classmethod
    def from_app(cls, client: Benchling, app_id: str) -> DependencyLinkStore:
        """
        From App.

        Instantiate a DependencyLinkStore from an app_id and a configured Benchling instance. Preferred to
        using the class's constructor.
        """
        config_provider = BenchlingConfigProvider(client, app_id)
        return DependencyLinkStore(config_provider)

    @property
    def configuration(self) -> List[ConfigurationReference]:
        """
        Get the underlying configuration.

        Return the raw, stored configuration. Can be used if the provided accessors are inadequate
        to find particular configuration items.
        """
        if not self._configuration:
            self._configuration = self._configuration_provider.config()
        return self._configuration

    @property
    def configuration_path_map(self) -> Dict[ConfigItemPath, ConfigurationReference]:
        """
        Config links.

        Return a map of configuration item paths to their corresponding configuration items.
        """
        if not self._configuration_map:
            self._configuration_map = {tuple(item.path): item for item in self.configuration}
        return self._configuration_map

    def config_by_path(self, path: List[str]) -> Optional[ConfigurationReference]:
        """
        Config by path.

        Find an app config item by its exact path match, if it exists. Does not search partial paths.
        """
        # Since we eager load all config now, we know that missing path means it's not configured in Benchling
        # Later if we support lazy loading, we'll need to differentiate what's in our cache versus missing
        return self.configuration_path_map.get(tuple(path))

    def invalidate_cache(self) -> None:
        """
        Invalidate Cache.

        Will force retrieval of configuration from the ConfigProvider the next time the link store is accessed.
        """
        self._configuration = None
        self._configuration_map = None


class HasAppConfigItem(Protocol):
    """
    Has App Config Item.

    A mixin for typing to assert that a class has an optional app config item attribute.
    """

    @property
    def path(self) -> str:
        """Return the path requested by the manifest."""
        ...

    @property
    def config_item(self) -> Optional[ConfigurationReference]:
        """Return the underlying app config item, if present."""
        ...


class HasScalarDefinition(Protocol):
    """
    Has Scalar Definition.

    A mixin for typing to assert that a particular class has scalar attributes.
    """

    @property
    def path(self) -> str:
        """Return the path requested by the manifest."""
        ...

    @property
    def config_item(self) -> Optional[ConfigurationReference]:
        """Return the underlying app config item, if present."""
        ...

    @property
    def definition(self) -> Optional[ScalarDefinition]:
        """Return the scalar definition, allowing for conversion to Python types."""
        ...


class HasConfigWithDecryptionProvider(Protocol):
    """
    Has Config With Decryption Provider.

    A mixin for typing to assert that a particular class has a decryption provider and config.
    """

    @property
    def path(self) -> str:
        """Return the path requested by the manifest."""
        ...

    @property
    def config_item(self) -> Optional[ConfigurationReference]:
        """Return the underlying app config item, if present."""
        ...

    @property
    def decryption_provider(self) -> Optional[BaseDecryptionProvider]:
        """Return the decryption provider."""
        ...


class RequiredValueMixin:
    """
    Required Value Mixin.

    A mixin for accessing a value which is required and should always be present. Should
    only be mixed in with HasAppConfigItem or another class that provides the `self.config_item` attribute.
    """

    @property
    def value(self: HasAppConfigItem) -> str:
        """Return the value of the app config item."""
        assert (
            self.config_item is not None and self.config_item.value is not None
        ), f"The app config item {self.path} is not set in Benchling"
        return str(self.config_item.value)


class RequiredScalarDependencyMixin(Generic[ScalarType]):
    """
    Require Scalar Config.

    A mixin for accessing a scalar config which is required and should always be present.
    Should only be mixed in with ScalarConfig.
    """

    @property
    def value(self: HasScalarDefinition) -> ScalarType:
        """Return the value of the scalar."""
        if self.definition:
            assert (
                self.config_item is not None and self.config_item.value is not None
            ), f"The app config item {self.path} is not set in Benchling"
            optional_typed_value = self.definition.from_str(value=str(self.config_item.value))
            assert optional_typed_value is not None
            return optional_typed_value
        raise MissingScalarDefinitionError(f"No definition registered for scalar config {self.path}")

    @property
    def value_str(self: HasScalarDefinition) -> str:
        """Return the value of the scalar as a string."""
        assert (
            self.config_item is not None and self.config_item.value is not None
        ), f"The app config item {self.path} is not set in Benchling"
        # Booleans are currently specified as str in the spec but are bool at runtime in JSON
        return str(self.config_item.value)


class RequiredSecureTextDependencyMixin(RequiredScalarDependencyMixin[str]):
    """
    Require Secure Text.

    A mixin for accessing a secure text config which is required and should always be present.
    Should only be mixed in with SecureTextConfig.
    """

    def decrypted_value(self: HasConfigWithDecryptionProvider) -> str:
        """
        Decrypted value.

        Decrypts a secure_text dependency's encrypted value into plain text.
        """
        assert (
            self.config_item is not None and self.config_item.value is not None
        ), f"The app config item {self.path} is not set in Benchling"
        assert (
            self.decryption_provider is not None
        ), f"The app config item {self.config_item} cannot be decrypted because no DecryptionProvider was set"
        return self.decryption_provider.decrypt(str(self.config_item.value))


@dataclass
class BaseConfigItem:
    """
    Base Config Item.

    The shared set of properties for all config items.
    """

    config_item: Optional[ConfigurationReference]


@dataclass
class ParentConfigItem(BaseConfigItem):
    """
    Parent Config Item.

    A parent config item which may have subdependencies that require context.
    """

    config_context: BaseDependencies


@dataclass
class ConfigItemContainer:
    """
    Config Item Container.

    Does not correspond to a linked app config item. Holds context passed from linked items.
    """

    parent_config: ParentConfigItem


@dataclass
class ScalarConfigItem(BaseConfigItem):
    """
    Scalar Config Item.

    Scalars are values that can be represented outside the Benchling domain.
    """

    definition: Optional[ScalarDefinition]


@dataclass
class SecureTextDependency(ScalarConfigItem):
    """
    SecureText Config.

    A dependency for accessing a secure_text config.
    """

    # This is declared Optional because a decryption provider is not required until attempting
    # to decrypt a value.
    decryption_provider: Optional[BaseDecryptionProvider]


class BaseDependencies:
    """
    A base class for implementing dependencies.

    Used as a facade for the underlying link store, which holds dependency links configured in Benchling.
    """

    _store: DependencyLinkStore
    _scalar_definitions: Dict[ScalarConfigItemType, ScalarDefinition]
    _unknown_scalar_definition: Optional[ScalarDefinition]
    # Will be required at runtime if an app attempts to decrypt a secure_text config
    _decryption_provider: Optional[BaseDecryptionProvider]

    def __init__(
        self,
        store: DependencyLinkStore,
        scalar_definitions: Dict[ScalarConfigItemType, ScalarDefinition] = DEFAULT_SCALAR_DEFINITIONS,
        unknown_scalar_definition: Optional[ScalarDefinition] = None,
        decryption_provider: Optional[BaseDecryptionProvider] = None,
    ):
        """
        Initialize Base Dependencies.

        :param store: The dependency link store to source dependency links from.
        :param scalar_definitions: A map of scalar types from the API definitions to ScalarDefinitions which
        determines how we want map them to concrete Python types and values. Can be overridden to customize
        deserialization behavior or formatting.
        :param unknown_scalar_definition: A scalar definition for handling unknown scalar types from the API. Can be
        used to control behavior for forwards compatibility with new types the SDK does not yet support (e.g.,
        by treating them as strings).
        :param decryption_provider: A decryption provider that can decrypt secrets from app config. If
        dependencies attempt to use a secure_text's decrypted value, a decryption_provider must be specified.
        """
        self._store = store
        self._scalar_definitions = scalar_definitions
        self._unknown_scalar_definition = unknown_scalar_definition
        self._decryption_provider = decryption_provider

    @classmethod
    def from_app(
        cls: Type[D],
        client: Benchling,
        app_id: str,
        decryption_provider: Optional[BaseDecryptionProvider] = None,
    ) -> D:
        """Initialize dependencies from an app_id."""
        link_store = DependencyLinkStore.from_app(client=client, app_id=app_id)
        return cls(link_store, decryption_provider=decryption_provider)

    def invalidate_cache(self) -> None:
        """Invalidate the cache of dependency links and force an update."""
        self._store.invalidate_cache()


def _supported_config_item(config_item: AppConfigItem) -> ConfigurationReference:
    if isinstance(config_item, UnknownType):
        raise UnsupportedDependencyError(
            f"Unable to read app configuration with unsupported type: {config_item}"
        )
    return config_item



from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.teams import get_team, list_teams
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import ListTeamsSort, Team, TeamsPaginatedList
from benchling_sdk.services.v2.base_service import BaseService


class TeamService(BaseService):
    """
    Teams.

    View team objects.

    See https://benchling.com/api/reference#/Teams
    """

    @api_method
    def get_by_id(self, team_id: str) -> Team:
        """
        Get a team by ID.

        See https://benchling.com/api/reference#/Teams/getTeam
        """
        response = get_team.sync_detailed(client=self.client, team_id=team_id)
        return model_from_detailed(response)

    @api_method
    def _teams_page(
        self,
        *,
        ids: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        mentioned_in: Optional[Iterable[str]] = None,
        organization_id: Optional[str] = None,
        has_members: Optional[Iterable[str]] = None,
        has_admins: Optional[Iterable[str]] = None,
        sort: Optional[ListTeamsSort] = None,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
    ) -> Response[TeamsPaginatedList]:
        response = list_teams.sync_detailed(
            client=self.client,
            ids=none_as_unset(optional_array_query_param(ids)),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            modified_at=none_as_unset(modified_at),
            mentioned_in=none_as_unset(optional_array_query_param(mentioned_in)),
            organization_id=none_as_unset(organization_id),
            has_members=none_as_unset(optional_array_query_param(has_members)),
            has_admins=none_as_unset(optional_array_query_param(has_admins)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            sort=none_as_unset(sort),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        *,
        ids: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        mentioned_in: Optional[Iterable[str]] = None,
        organization_id: Optional[str] = None,
        has_members: Optional[Iterable[str]] = None,
        has_admins: Optional[Iterable[str]] = None,
        sort: Optional[ListTeamsSort] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[Team]:
        """
        List teams.

        Returns all teams that the caller has permission to view. The following roles have view permission:
        * tenant admins
        * members of the team's organization

        See https://benchling.com/api/reference#/Teams/listTeams
        """

        def api_call(next_token: NextToken) -> Response[TeamsPaginatedList]:
            return self._teams_page(
                ids=ids,
                name=name,
                name_includes=name_includes,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                modified_at=modified_at,
                mentioned_in=mentioned_in,
                organization_id=organization_id,
                has_members=has_members,
                has_admins=has_admins,
                sort=sort,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: TeamsPaginatedList) -> Optional[List[Team]]:
            return body.teams

        return PageIterator(api_call, results_extractor)



"""Specialized Exception classes."""
from dataclasses import dataclass
import json
from json import JSONDecodeError
from typing import Any, Dict, MutableMapping, Optional, Union

from benchling_api_client.v2.types import Response

from benchling_sdk.helpers.logging_helpers import default_logger
from benchling_sdk.helpers.serialization_helpers import unset_as_none
from benchling_sdk.models import (
    AsyncTask,
    AsyncTaskStatus,
    BadRequestError,
    BadRequestErrorBulk,
    ConflictError,
    ForbiddenError,
    NotFoundError,
)

logger = default_logger()


@dataclass
class BenchlingError(Exception):
    """An error resulting from communicating with the Benchling API.

    This could be an error returned from the API intentionally (e.g., 400 Bad Request) or an
    unexpected transport error (e.g., 502 Bad Gateway)

    The json attribute is present if the API response provided a deserializable JSON body as part of the
    error description. It will be None if the response could not be parsed as JSON.

    The content attribute is any unparsed content returned as part of the response body.
    """

    status_code: int
    headers: MutableMapping[str, str]
    json: Optional[Dict[str, str]]
    content: Optional[bytes]
    parsed: Union[
        None,
        ForbiddenError,
        NotFoundError,
        BadRequestError,
        BadRequestErrorBulk,
        ConflictError,
    ]

    @classmethod
    def from_response(cls, response: Response) -> "BenchlingError":
        """Create a BenchlingError from a generated Response."""
        json_body = _parse_error_body(response.content)
        return cls(
            status_code=response.status_code,
            headers=response.headers,
            json=json_body,
            content=response.content,
            parsed=response.parsed,
        )

    def __str__(self):
        message = self.json if self.json else self.content
        return f"{self.__class__.__name__}(status_code={self.status_code}, message={message})"

    def __hash__(self):
        return self._generate_hash()

    def __eq__(self, other):
        return self._generate_hash() == other._generate_hash()

    def _generate_hash(self):
        return hash(
            (
                self.status_code,
                json.dumps(self.headers, sort_keys=True),
                json.dumps(self.json, sort_keys=True),
                self.content,
                str(self.parsed),
            )
        )


@dataclass
class RegistrationError(Exception):
    """An error relating to Benchling registration."""

    message: Optional[str] = None
    errors: Optional[Dict[Any, Any]] = None
    task_status: Optional[AsyncTaskStatus] = None

    @classmethod
    def from_task(cls, task: AsyncTask) -> "RegistrationError":
        """Create a RegistrationError from a failed AsyncTask."""
        task_errors = unset_as_none(task.errors)
        errors_dict: Dict[Any, Any] = task_errors.to_dict() if task_errors else dict()  # type: ignore
        return cls(message=unset_as_none(task.message), errors=errors_dict, task_status=task.status)

    def __str__(self):
        return repr(self)


@dataclass
class WaitForTaskExpiredError(Exception):
    """An error indicating an AsyncTask did not complete in the time allotted for polling."""

    message: str
    task: AsyncTask

    def __str__(self):
        return repr(self)


def raise_for_status(response: Response) -> None:
    """Evaluate a Response for a successful HTTP status code or raise a BenchlingError."""
    logger.info("Status: %s", response.status_code)
    logger.debug(response)
    if response.status_code < 200 or response.status_code >= 300:
        raise BenchlingError.from_response(response)


def _parse_error_body(content: bytes) -> Optional[Dict[str, str]]:
    if content:
        string_content = content.decode("utf-8")
        # In case the response is not a serialized JSON dict (e.g. a gateway error)
        try:
            return json.loads(string_content)
        except JSONDecodeError:
            # Some responses may not be JSON. Just catch the error
            return None
    return None
